import sys
import csv
import _csv
from io import TextIOWrapper
from typing import Optional
from nomenklatura.entity import CE
from nomenklatura.statement.serialize import pack_statement, PACK_COLUMNS
from followthemoney.util import PathLike

from zavod.sinks.common import FileSink


class PackStatementSink(FileSink[CE]):
    def __init__(self, path: PathLike) -> None:
        super().__init__(path)
        self.wrapper: Optional[TextIOWrapper] = None
        self.writer: Optional[_csv._writer] = None

    def emit(self, entity: CE) -> None:
        with self.lock:
            if self.fh is None:
                self.fh = open(self.path, "wb")
            if self.wrapper is None and self.fh is not None:
                self.wrapper = TextIOWrapper(self.fh, encoding="utf-8")
            if self.writer is None and self.wrapper is not None:
                self.writer = csv.writer(
                    self.wrapper,
                    dialect=csv.unix_dialect,
                    quoting=csv.QUOTE_MINIMAL,
                )
            if self.writer is not None:
                for stmt in entity.statements:
                    row = pack_statement(stmt)
                    self.writer.writerow([row.get(c) for c in PACK_COLUMNS])

    def close(self) -> None:
        with self.lock:
            if self.wrapper is not None:
                self.wrapper.close()
                self.wrapper = None
            super().close()

    def __repr__(self) -> str:
        return f"<PackStatementSink({self.path!r})>"



from zavod.sinks.common import Sink
from zavod.sinks.json_entity import JSONEntitySink
from zavod.sinks.json_statement import JSONStatementSink
from zavod.sinks.pack_statement import PackStatementSink

__all__ = [
    "Sink",
    "JSONEntitySink",
    "JSONStatementSink",
    "PackStatementSink",
]



import sys
import logging
import structlog
from typing import List, Any, Dict, MutableMapping
from structlog.stdlib import get_logger as get_raw_logger
from structlog.contextvars import merge_contextvars
from structlog.types import Processor

from zavod import settings


def configure_logging(
    level: int = logging.DEBUG,
    extra_processors: List[Processor] = [],
) -> None:
    """Configure log levels and structured logging."""
    processors: List[Processor] = [
        structlog.stdlib.add_log_level,
        structlog.stdlib.add_logger_name,
        structlog.processors.StackInfoRenderer(),
        structlog.processors.format_exc_info,
        merge_contextvars,
        structlog.dev.set_exc_info,
        structlog.processors.UnicodeDecoder(),
    ]
    processors.extend(extra_processors)

    if settings.LOG_JSON:
        processors.append(structlog.processors.TimeStamper(fmt="iso"))
        processors.append(format_json)
        formatter = structlog.stdlib.ProcessorFormatter(
            foreign_pre_chain=processors,
            processor=structlog.processors.JSONRenderer(),
        )
    else:
        processors.append(structlog.processors.TimeStamper(fmt="%Y-%m-%d %H:%M:%S"))
        formatter = structlog.stdlib.ProcessorFormatter(
            foreign_pre_chain=processors,
            processor=structlog.dev.ConsoleRenderer(
                exception_formatter=structlog.dev.plain_traceback
            ),
        )

    all_processors = processors + [
        structlog.stdlib.ProcessorFormatter.wrap_for_formatter
    ]

    # configuration for structlog based loggers
    structlog.configure(
        cache_logger_on_first_use=True,
        wrapper_class=structlog.stdlib.BoundLogger,
        processors=all_processors,
        logger_factory=structlog.stdlib.LoggerFactory(),
    )

    handler = logging.StreamHandler(sys.stderr)
    handler.setLevel(level)
    handler.setFormatter(formatter)

    logger = logging.getLogger()
    logger.setLevel(level)
    logger.addHandler(handler)


def get_logger(name: str) -> structlog.stdlib.BoundLogger:
    return get_raw_logger(name)


def format_json(
    _: Any, __: str, ed: MutableMapping[str, str]
) -> MutableMapping[str, str]:
    """Stackdriver uses `message` and `severity` keys to display logs"""
    ed["message"] = ed.pop("event")
    ed["severity"] = ed.pop("level", "info").upper()
    return ed



import sys
from threading import RLock
from typing import Generic, BinaryIO, Optional
from followthemoney.util import PathLike

from nomenklatura.entity import CE


class Sink(Generic[CE]):
    def __init__(self, path: PathLike) -> None:
        self.path = path

    def emit(self, entity: CE) -> None:
        raise NotImplemented

    def close(self) -> None:
        pass


class FileSink(Sink[CE]):
    def __init__(self, path: PathLike) -> None:
        super().__init__(path)
        self.lock = RLock()
        self.fh: Optional[BinaryIO] = None

    def emit_locked(self, fh: BinaryIO, entity: CE) -> None:
        raise NotImplemented

    def emit(self, entity: CE) -> None:
        with self.lock:
            if self.fh is None:
                self.fh = open(self.path, "wb")
            self.emit_locked(self.fh, entity)

    def close(self) -> None:
        with self.lock:
            if self.fh is not None:
                self.fh.close()
                self.fh = None

    def __str__(self) -> str:
        return str(self.path)



from typing import Dict, Any, TypeVar
from normality import slugify
from nomenklatura.dataset import Dataset, DataCatalog
from nomenklatura.util import datetime_iso

from zavod import settings

ZD = TypeVar("ZD", bound="ZavodDataset")


class ZavodDataset(Dataset):
    def __init__(self, catalog: DataCatalog[ZD], data: Dict[str, Any]):
        super().__init__(catalog, data)
        self.prefix: str = data.get("prefix", slugify(self.name, sep="-"))
        if self.updated_at is None:
            self.updated_at = datetime_iso(settings.RUN_TIME)



from zavod.parse.addresses import make_address, format_address
from zavod.parse.xml import remove_namespace
from zavod.parse.names import make_name, apply_name

__all__ = [
    "make_address",
    "format_address",
    "remove_namespace",
    "make_name",
    "apply_name",
]



from functools import cache, lru_cache
from typing import Optional

from addressformatting import AddressFormatter
from followthemoney.types import registry
from followthemoney.util import join_text, make_entity_id
from nomenklatura.entity import CE
from normality import slugify
from zavod.context import GenericZavod
from zavod.dataset import ZD


@cache
def get_formatter() -> AddressFormatter:
    return AddressFormatter()


@lru_cache(maxsize=5000)
def format_address(
    summary: Optional[str] = None,
    po_box: Optional[str] = None,
    street: Optional[str] = None,
    house: Optional[str] = None,
    house_number: Optional[str] = None,
    postal_code: Optional[str] = None,
    city: Optional[str] = None,
    county: Optional[str] = None,
    state: Optional[str] = None,
    country_code: Optional[str] = None,
) -> str:
    data = {
        "attention": summary,
        "house": po_box,
        "road": street,
        "house": house,
        "house_number": house_number,
        "postcode": postal_code,
        "city": city,
        "county": county,
        "state_district": state,
    }
    return get_formatter().one_line(data, country=country_code)


def make_address(
    context: GenericZavod[CE, ZD],
    full: Optional[str] = None,
    remarks: Optional[str] = None,
    summary: Optional[str] = None,
    po_box: Optional[str] = None,
    street: Optional[str] = None,
    street2: Optional[str] = None,
    street3: Optional[str] = None,
    city: Optional[str] = None,
    place: Optional[str] = None,
    postal_code: Optional[str] = None,
    state: Optional[str] = None,
    region: Optional[str] = None,
    country: Optional[str] = None,
    country_code: Optional[str] = None,
    key: Optional[str] = None,
    lang: Optional[str] = None,
) -> CE:
    """Generate an address schema object adjacent to the main entity."""

    city = join_text(place, city, sep=", ")
    street = join_text(street, street2, street3, sep=", ")

    address = context.make("Address")
    address.add("full", full, lang=lang)
    address.add("remarks", remarks, lang=lang)
    address.add("summary", summary, lang=lang)
    address.add("postOfficeBox", po_box, lang=lang)
    address.add("street", street, lang=lang)
    address.add("city", city, lang=lang)
    address.add("postalCode", postal_code, lang=lang)
    address.add("region", region, lang=lang)
    address.add("state", state, quiet=True, lang=lang)
    address.add("country", country, lang=lang)
    address.add("country", country_code)

    country_code = address.first("country")
    if not full:
        full = format_address(
            summary=summary,
            po_box=po_box,
            street=street,
            postal_code=postal_code,
            city=city,
            state=join_text(region, state, sep=", "),
            country_code=country_code,
        )

    full_country = registry.country.clean(full)
    if full_country is not None:
        address.add("country", full_country, lang=lang)
        # full = None

    # full = clean_address(full)
    address.add("full", full, lang=lang)

    if full:
        norm_full = slugify(full)
        hash_id = make_entity_id(country_code, norm_full, key)
        if hash_id is not None:
            address.id = f"addr-{hash_id}"
    return address



from typing import BinaryIO
from nomenklatura.entity import CE
from followthemoney.cli.util import write_entity

from zavod.sinks.common import FileSink


class JSONEntitySink(FileSink[CE]):
    def emit_locked(self, fh: BinaryIO, entity: CE) -> None:
        write_entity(fh, entity)

    def __repr__(self) -> str:
        return f"<JSONEntitySink({self.path!r})>"



import warnings
from typing import Any, Optional
from functools import partial
from pathlib import Path
from requests import Session
from urllib3.connectionpool import InsecureRequestWarning

from zavod import settings
from zavod.logs import get_logger

log = get_logger(__name__)
warnings.filterwarnings("ignore", category=InsecureRequestWarning)


def make_session(user_agent: str = settings.HTTP_USER_AGENT) -> Session:
    session = Session()
    session.headers["User-Agent"] = user_agent
    session.verify = False
    session.request = partial(session.request, timeout=settings.HTTP_TIMEOUT)  # type: ignore
    return session


def fetch_file(
    session: Session,
    url: str,
    name: str,
    data_path: Path = settings.DATA_PATH,
    auth: Optional[Any] = None,
    headers: Optional[Any] = None,
) -> Path:
    """Fetch a (large) file via HTTP to the data path."""
    out_path = data_path.joinpath(name)
    if out_path.exists():
        return out_path
    log.info("Fetching file", url=url)
    out_path.parent.mkdir(parents=True, exist_ok=True)
    with session.get(url, auth=auth, headers=headers, stream=True) as res:
        res.raise_for_status()
        with open(out_path, "wb") as fh:
            for chunk in res.iter_content(chunk_size=8192 * 10):
                fh.write(chunk)
    return out_path



import logging
from pathlib import Path
from typing import Generator, Optional, Type
from contextlib import contextmanager
from nomenklatura.entity import CompositeEntity
from followthemoney.util import PathLike

from zavod import settings
from zavod.context import GenericZavod
from zavod.dataset import ZavodDataset, ZD
from zavod.logs import configure_logging, get_logger
from zavod.sinks.common import Sink
from zavod.sinks.json_entity import JSONEntitySink

__version__ = "0.7.0"
__all__ = [
    "init",
    "context",
    "Zavod",
    "ZavodDataset",
    "ZD",
    "PathLike",
    "configure_logging",
    "get_logger",
    "settings",
]

logging.getLogger("prefixdate").setLevel(logging.ERROR)


class Zavod(GenericZavod[CompositeEntity, ZavodDataset]):
    pass


def init(
    metadata_path: PathLike,
    verbose: bool = False,
    data_path: Path = settings.DATA_PATH,
    out_file: Optional[PathLike] = "fragments.json",
    sink: Optional[Sink[CompositeEntity]] = None,
    sink_cls: Type[Sink[CompositeEntity]] = JSONEntitySink,
) -> Zavod:
    """Initiate the zavod working environment and create a processing context."""
    level = logging.DEBUG if verbose else logging.INFO
    configure_logging(level=level)
    if out_file is not None and sink is None:
        out_path = data_path.joinpath(out_file)
        out_path.parent.mkdir(exist_ok=True, parents=True)
        sink = sink_cls(out_path)
    dataset = ZavodDataset.from_path(metadata_path)
    return Zavod(dataset, CompositeEntity, data_path=data_path, sink=sink)


@contextmanager
def init_context(
    metadata_path: PathLike,
    verbose: bool = False,
    data_path: Path = settings.DATA_PATH,
    out_file: Optional[PathLike] = "fragments.json",
    sink: Optional[Sink[CompositeEntity]] = None,
) -> Generator[Zavod, None, None]:
    ctx = init(
        metadata_path,
        verbose=verbose,
        data_path=data_path,
        out_file=out_file,
        sink=sink,
    )
    try:
        yield ctx
    finally:
        ctx.close()



from typing import BinaryIO
from nomenklatura.entity import CE
from nomenklatura.statement.serialize import write_json_statement

from zavod.sinks.common import FileSink


class JSONStatementSink(FileSink[CE]):
    def emit_locked(self, fh: BinaryIO, entity: CE) -> None:
        for stmt in entity.statements:
            write_json_statement(fh, stmt)

    def __repr__(self) -> str:
        return f"<JSONStatementSink({self.path!r})>"
